#include <cmath>
#include <cstdlib>

#include <pico/stdio_usb.h>
#include <pico/stdlib.h>
#include <pico/time.h>
#include <hardware/pwm.h>
#include <hardware/i2c.h>

#include <picoled/ST7789.h>
#include <picoled/gfx/2D.h>


using namespace picoled;

namespace
{

float get_random()
{
  return rand() / static_cast<float>(RAND_MAX);
}

int32_t get_encoder_delta(i2c_inst* i2c, uint8_t address)
{
  // SEESAW_ENCODER_BASE and SEESAW_ENCODER_DELTA values
  // https://github.com/adafruit/Adafruit_Seesaw/blob/master/Adafruit_seesaw.h
  uint8_t buf[4] = { 0x11, 0x40, 0x00, 0x00 };
  i2c_write_blocking(i2c, address, buf, 2, false);
  sleep_us(250);
  i2c_read_blocking(i2c, address, buf, 4, false);
  return
    (static_cast<int32_t>(buf[0]) << 24) |
    (static_cast<int32_t>(buf[1]) << 16) |
    (static_cast<int32_t>(buf[2]) << 8 ) |
    (static_cast<int32_t>(buf[3]) << 0 );
}

constexpr uint8_t LEFT_ENCODER_ADDRESS = 0x36;
constexpr uint8_t RIGHT_ENCODER_ADDRESS = 0x37;
constexpr uint8_t OLED_ADDRESS = 0x3C;

}

int main()
{
  //set_sys_clock_khz(133000, false);

  stdio_usb_init();
  srand(time_us_64());

  i2c_init(i2c0, 1000 * 1000);
  gpio_set_function(0, GPIO_FUNC_I2C);
  gpio_set_function(1, GPIO_FUNC_I2C);
  gpio_pull_up(0);
  gpio_pull_up(1);

  spi_init(spi0, 32000000);
  gpio_set_function(2, GPIO_FUNC_SPI);
  gpio_set_function(3, GPIO_FUNC_SPI);
  gpio_set_function(4, GPIO_FUNC_SPI);

  gpio_set_function(5, GPIO_FUNC_SIO);
  gpio_set_dir(5, GPIO_OUT);
  gpio_put(5, false);

  gpio_set_function(6, GPIO_FUNC_SIO);
  gpio_set_dir(6, GPIO_OUT);
  gpio_put(6, false);

  st7789 oled(135, 240, spi0, 5, 6);
  gfx::gfx2d gfx(oled);

  float angle = 0;
  int64_t time = time_us_32();

  buffer<pixel> tex(64, 64);
  for (int x = 0; x < 64; x++)
  {
    for (int y = 0; y < 64; y++)
    {
      const uint8_t r = 4 * x;
      const uint8_t g = 4 * y;
      const uint8_t b = 254 - 2 * x - 2 * y;
      tex.write(x, y, { r, g, b });
    }
  }

  while (true)
  {
    int64_t current = time_us_32();
    const float dt = std::abs(current - time) / 1000000.0f;
    time = current;

    stdio_printf("%05f\n", dt);

    const int32_t delta = get_encoder_delta(i2c0, RIGHT_ENCODER_ADDRESS);
    angle -= delta / 12.0f * 3.1415926f;

    const float angle2 = angle + (2 * 3.1415926f) / 3;
    const float angle3 = angle + 2 * (2 * 3.1415926f) / 3;

    const gfx::vec2 ap = { 60.0f + 40 * cosf(angle), 60.0f + 40 * sinf(angle) };
    const gfx::vec2 bp = { 60.0f + 40 * cosf(angle2), 60.0f + 40 * sinf(angle2) };
    const gfx::vec2 cp = { 60.0f + 40 * cosf(angle3), 60.0f + 40 * sinf(angle3) };

    const gfx::vec2 at = { 0.0f, 0.0f };
    const gfx::vec2 bt = { 1.0f, 0.0f };
    const gfx::vec2 ct = { 0.0f, 1.0f };

    oled.clear();
    gfx.draw_triangle({
      { ap, at },
      { bp, bt },
      { cp, ct }
    }, tex);

    oled.swap_buffers();
    oled.update();
  }
}
